@@ -1,7 +1,23 @@
 #!/usr/bin/env python
 
 """Module containing wrapper classes over paramiko.SSHClient
-See SSHClient and ParallelSSHClient"""
+See SSHClient and ParallelSSHClient
+
+Copyright (C) 2014 Panos Kittenis
+
+This library is free software; you can redistribute it and/or
+modify it under the terms of the GNU Lesser General Public
+License as published by the Free Software Foundation, version 2.1.
+
+This library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+Lesser General Public License for more details.
+
+You should have received a copy of the GNU Lesser General Public
+License along with this library; if not, write to the Free Software
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+"""
 
 import gevent.pool
 from gevent import socket
@@ -70,9 +86,9 @@ class SSHClient(object):
         :type port: int
         :param pkey: (Optional) Client's private key to be used to connect with
         :type pkey: :mod:`paramiko.PKey`
-        :raises: AuthenticationException on authentication error
-        :raises: UnknownHostException on DNS resolution error
-        :raises: ConnectionErrorException on error connecting"""
+        :raises: :mod:`pssh.AuthenticationException` on authentication error
+        :raises: :mod:`pssh.UnknownHostException` on DNS resolution error
+        :raises: :mod:`pssh.ConnectionErrorException` on error connecting"""
         ssh_config = paramiko.SSHConfig()
         _ssh_config_file = os.path.sep.join([os.path.expanduser('~'),
                                              '.ssh',
@@ -147,6 +163,7 @@ class SSHClient(object):
         """
         channel = self.client.get_transport().open_session()
         channel.get_pty()
+        # stdin (unused), stdout, stderr
         (_, stdout, stderr) = (channel.makefile('wb'), channel.makefile('rb'),
                                channel.makefile_stderr('rb'))
         if sudo:
@@ -213,7 +230,11 @@ class SSHClient(object):
 
 class ParallelSSHClient(object):
     """Uses :mod:`pssh.SSHClient`, performs tasks over SSH on multiple hosts in \
-    parallel"""
+    parallel.
+
+    Connections to hosts are established in parallel when an object of this
+    class is created, therefor any connection and/or authentication exceptions \
+    will happen on creation and need to be caught."""
 
     def __init__(self, hosts,
                  user=None, password=None, port=None, pkey=None,
@@ -221,8 +242,6 @@ class ParallelSSHClient(object):
         """
         :param hosts: Hosts to connect to
         :type hosts: list(str)
-        :param pool_size: Pool size - how many commands to run in parallel
-        :type pool_size: int
         :param user: (Optional) User to login as. Defaults to logged in user or\
         user from ~/.ssh/config or /etc/ssh/ssh_config if set
         :type user: str
@@ -237,9 +256,9 @@ class ParallelSSHClient(object):
         :param pool_size: (Optional) Greenlet pool size. Controls on how many\
         hosts to execute tasks in parallel. Defaults to 10
         :type pool_size: int
-        :raises: AuthenticationException on authentication error
-        :raises: UnknownHostException on DNS resolution error
-        :raises: ConnectionErrorException on error connecting
+        :raises: :mod:`pssh.AuthenticationException` on authentication error
+        :raises: :mod:`pssh.UnknownHostException` on DNS resolution error
+        :raises: :mod:`pssh.ConnectionErrorException` on error connecting
         
         **Example**
 
@@ -256,6 +275,30 @@ class ParallelSSHClient(object):
         >>> print output
         [{'myhost1': {'exit_code': 2}}, {'myhost2': {'exit_code': 2}}]
 
+        **Example with returned stdout and stderr buffers**
+
+        >>> from pssh import ParallelSSHClient, AuthenticationException,\
+        		UnknownHostException, ConnectionErrorException
+        >>> try:
+        >>> ... client = ParallelSSHClient(['myhost1', 'myhost2'])
+        >>> except (AuthenticationException, UnknownHostException, ConnectionErrorException):
+        >>> ... return
+        >>> cmds = client.exec_command('ls -ltrh /tmp/aasdfasdf', sudo = True)
+        >>> output = [client.get_stdout(cmd, return_buffers=True) for cmd in cmds]
+        >>> print output
+        [{'myhost1': {'exit_code': 2,
+        	      'stdout' : <generator object <genexpr>,
+                      'stderr' : <generator object <genexpr>,}},
+         {'myhost2': {'exit_code': 2,
+         	      'stdout' : <generator object <genexpr>,
+                      'stderr' : <generator object <genexpr>,}},
+                      ]
+        >>> for host_stdout in output:
+            ... for line in host_stdout[host_output.keys()[0]]['stdout']:
+                ... print line
+                    ls: cannot access /tmp/aasdfasdf: No such file or directory
+                    ls: cannot access /tmp/aasdfasdf: No such file or directory
+
         **Example with specified private key**
 
         >>> import paramiko
@@ -331,29 +374,47 @@ class ParallelSSHClient(object):
                                                 port=self.port, pkey=self.pkey)
         return self.host_clients[host].exec_command(*args, **kwargs)
 
-    def get_stdout(self, greenlet):
-        """Print stdout from greenlet and return exit code for host
-        :param greenlet: Greenlet object containing an \
-        SSH channel reference, hostname, stdout and stderr buffers
-        :type greenlet: :mod:`gevent.Greenlet`
-
-        :mod:`pssh.get_stdout` will close the open SSH channel but this does **not**
-        close the established connection to the remote host, only the
+    def get_stdout(self, greenlet, return_buffers=False):
+        """Get/print stdout from greenlet and return exit code for host
+        
+        :mod:`pssh.get_stdout` will close the open SSH channel but this does
+        **not** close the established connection to the remote host, only the
         authenticated SSH channel within it. This is standard practise
         in SSH when a command has finished executing. A new command
         will open a new channel which is very fast on already established
         connections.
 
-        :rtype: Dictionary containing {host: {'exit_code': exit code}} entry \
-        for example {'myhost1': {'exit_code': 0}}
+        By default, stdout and stderr will be logged via the logger named \
+        ``host_logger`` unless ``return_buffers`` is set to ``True`` in which case \
+        both buffers are instead returned along with the exit status.
+
+        :param greenlet: Greenlet object containing an \
+        SSH channel reference, hostname, stdout and stderr buffers
+        :type greenlet: :mod:`gevent.Greenlet`
+
+        :param return_buffers: Flag to turn on returning stdout and stderr \
+        buffers along with exit code. Defaults to off.
+        :type return_buffers: bool
+
+        :rtype: Dictionary containing ``{host: {'exit_code': exit code}}`` entry \
+        for example ``{'myhost1': {'exit_code': 0}}``
+        :rtype: With ``return_buffers=True``: ``{'myhost1': {'exit_code': 0,
+         						     'stdout' : <iterable>,
+                                                             'stderr' : <iterable>,}}``
         """
-        channel, host, stdout, stderr = greenlet.get()
-        for line in stdout:
-            host_logger.info("[%s]\t%s", host, line.strip(),)
-        for line in stderr:
-            host_logger.info("[%s] [err] %s", host, line.strip(),)
+        channel, host, _stdout, _stderr = greenlet.get()
+        stdout = (line.strip() for line in _stdout)
+        stderr = (line.strip() for line in _stderr)
         channel.close()
-        return {host: {'exit_code': channel.recv_exit_status()}}
+        if not return_buffers:
+            for line in stdout:
+                host_logger.info("[%s]\t%s", host, line,)
+            for line in stderr:
+                host_logger.info("[%s] [err] %s", host, line,)
+            return {host: {'exit_code': channel.recv_exit_status(),}}
+        return {host: {'exit_code': channel.recv_exit_status(),
+                       'stdout' : stdout,
+                       'stderr' : stderr, }}
 
     def copy_file(self, local_file, remote_file):
         """Copy local file to remote file in parallel
@@ -397,7 +458,8 @@ def test():
 def test_parallel():
     client = ParallelSSHClient(['localhost'])
     cmds = client.exec_command('ls -ltrh')
-    print [client.get_stdout(cmd) for cmd in cmds]
+    output = [client.get_stdout(cmd, return_buffers=True) for cmd in cmds]
+    print output
     cmds = client.copy_file('../test', 'test_dir/test')
     client.pool.join()
 
